import sys

def read_ref(refseq_ref_file):
    ref_dict = {}
    for line in file(refseq_ref_file):
        ls = line.strip().split(' ')
        refseq = ls[0]
        genename = ls[1]
        chromosome = ls[2]
        start = ls[3]
        end = ls[4]
        strand = '+' if ls[5] == '1' else '-'
        
        ref_dict[refseq] = {'genename': genename, 'chromosome': chromosome,
                    'start': start, 'end': end, 'strand': strand}
    return ref_dict
    
def main(em_output_file, refseq_ref_file, exons_file):
    ref_dict = read_ref(refseq_ref_file)
    
    count = 0
    count_dict = {}
    for line in file(exons_file):
        count += 1
        ls = line.strip().split('\t')
        split_point = int(ls[1])
        running_sum = 0
        border = []
        for i in range(2, len(ls)):
            running_sum += int(ls[i].split('_')[-1]) - int(ls[i].split('_')[-2])
            if running_sum == split_point:
                border.append( (ls[i].split('_')[0], ls[i].split('_')[-1]) )
                border.append( (ls[i+1].split('_')[0], ls[i+1].split('_')[-2]) )
                count_dict[count] = border
                break
    
    for line in file(em_output_file):
        ls = line.strip().split('\t')
        US_refseq = ls[0].split('.')[0][2:]
        DS_refseq = ls[0].split('.')[1][:ls[0].split('.')[1].find('_', 3)]
        fusion_number = int(ls[0].split('_')[-1])
        US_dict = ref_dict[US_refseq]
        DS_dict = ref_dict[DS_refseq]
        sys.stdout.write(US_dict['chromosome'] + '\t' + US_dict['start'] + '\t')
        sys.stdout.write(count_dict[fusion_number][0][1] + '\t')
        sys.stdout.write(DS_dict['chromosome'] + '\t')
        sys.stdout.write(count_dict[fusion_number][1][1] + '\t')
        sys.stdout.write(DS_dict['end'] + '\t' + ls[0] + '\t' + ls[1] + '\t')
        sys.stdout.write(US_dict['strand'] + '\t' + DS_dict['strand'] + '\t')
        sys.stdout.write(US_dict['genename'] + '\t' + DS_dict['genename'] + '\n')
        
if __name__ == '__main__':
    main(*sys.argv[1:])