#!/usr/bin/python
import sys

TRANSCRIPT_TAG = 'ucsc'

class tx_regions(object):

    def __init__(self, txfile, fuzz = 1500):

        self._FAIL_MULTIPLE = 100
        self._chrom_dict = {}
        self._fuzz = int(fuzz)
        for line in file(txfile):
            ls = line.strip('\n').split('\t')
            self._chrom_dict.setdefault(ls[0], []).append( (int(ls[1]), int(ls[2])) )

        for chrom in self._chrom_dict:
            self._chrom_dict[chrom].sort(key=lambda x:x[0])

    def get_tx_range(self, chrom, coordinate):
        if chrom not in self._chrom_dict:
            return [coordinate - 20*self._fuzz, coordinate + 20*self._fuzz]

        for region in self._chrom_dict[chrom]:
            if region[1] < coordinate:
                continue
            elif region[0] <= coordinate:
                return [region[0] - self._fuzz, region[1] + self._fuzz]
            else:
                return [coordinate - self._FAIL_MULTIPLE*self._fuzz, coordinate + self._FAIL_MULTIPLE*self._fuzz]

        return [coordinate - self._FAIL_MULTIPLE*self._fuzz, coordinate + self._FAIL_MULTIPLE*self._fuzz]

class bt_iterator(object):

    def __init__(self, pslfile):
        self._file = open(pslfile, 'r')

    def __iter__(self):
        return self

    def next(self):
        nl = self._file.readline()

        if len(nl) == 0:
            return None

        nl_list = nl.strip("\n").split("\t")

        read_id = nl_list[0]
        mapping = nl_list[2]
        mapping_start = int(nl_list[3])
        read_length = len(nl_list[4])

        if TRANSCRIPT_TAG in mapping:
            transcript_coordinate = int(mapping.split('_')[-1])
            mapping_start += transcript_coordinate
            mapping = '_'.join(mapping.split('_')[2:-1])

        return {'read_id':read_id, 'mapping':mapping, 'mapping_start':mapping_start,
                'read_length':read_length}



def main(bt_results_file, tx_regions_file, fq_bust_file, fuzz = '1500'):


    for line in file(fq_bust_file):
        if line[0] == '@':
            #So, this might lead to false positives depending on the
            #encoding of quality scores, but a few false positive
            #aren't a problem here.
            sys.stdout.write(line[1:-3] + '\n')
            
    bi = bt_iterator(bt_results_file)
    txf = tx_regions(tx_regions_file, fuzz)


    read_alns_1 = {} ##Dictionary from read_id to list of hits
    read_alns_2 = {} ##Same for other paired end

    counter = 0
    while 1:

        rec = bi.next()

        if rec is None:
            break

        counter += 1

        read_id = rec['read_id']

        ##This is kind of specific to the format of your read ids
        ##end shoudl be '1' or '2'
        ##read_id should end with '/1' or '/2'

        end = read_id[-1]
        base_read_id = read_id[:-2]
        read_length = rec['read_length']

        if base_read_id not in eval('read_alns_' + end):
            eval('read_alns_' + end)[base_read_id] = []


        mapping = rec['mapping']

        eval('read_alns_' + end)[base_read_id].append((mapping, rec['mapping_start']))

    counter = 0

    for read in read_alns_1:
        maps1 = read_alns_1[read]

        try:
            maps2 = read_alns_2[read]
        except KeyError:
            continue

        concordant = False

        concordant = testmaps(maps1, maps2, txf)
        if concordant:
           print read

def testmaps(maps1, maps2, txf):
    for map1 in maps1:
        for map2 in maps2:
            if map1[0] == map2[0]:
                map1range = txf.get_tx_range(map1[0], map1[1])
                map2range = txf.get_tx_range(map2[0], map2[1])
                if map2range[0] <= map1[1] <= map2range[1]  or \
                    map1range[0] <= map2[1] <= map1range[1]:
                    return True
    return False
    
    
if __name__ == '__main__':
    main(*sys.argv[1:])

