//C++ include
#include <map>
#include <string>
//C includes
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//Samtools
#include "bam.h"

/* Some globals. */
std::map<std::string, int> isoform_to_column
std::map<std::string, std::vector<std::string> > refseq_to_fusion;
std::map<std::string, int> isoform_lengths;
std::vector<unsigned int> M_rows;
std::vector<unsigned int> M_cols;
std::vector<double> M_values;
std::vector<double> read_sums;
std::vector<double> theta;
std::vector<double> distance_probabilites;
std::vector<unsigned int> header1_lengths;
std::vector<unsigned int> fusion_lengths;

double M_value(bam1_t* b1, bam1_t* b2, bool fusion)
{
    double ret_val = 1;
    int m_distance = abs(b1->core.pos - b2->core.pos);
    if(m_distance < 25000)
    {
        ret_val *= distance_probabilities[i];
    }
    else
    {
        return 0;
    }

    int iso_len;
    if(fusion)
    {
        iso_len = fusion_lengths[b1->core.tid];
    }
    else
    {
        iso_len = header1_lengths[b1->score.tid];
    }

    ret_val /= (double)iso_len;

    bam1_t* mcur;
    mcur = bam_init1();

    int i,j;
    for(i=0; i<2; i++)
    {
        if(i==0)
        {
            bam_copy1(mcur, b1);
        }
        else
        {
            bam_copy1(mcur, b2);
        }

        double Pcur = 1;

        /* First get the mismatch indices from the MD tag */
        std::set<short int> mismatches;
        char* MDtag = bam_aux_get(mcur, "MD");
        int offset = 0;
        std::string digit_string = "";
        for(j=1; j<strlen(MDtag); j++) /* The first char is always i */
        {
            if(isalpha(MDtag[j]))
            {
                if(digit_string.length() > 0)
                {
                    offset += atoi(digit_string.c_str());
                    digit_string = "";
                }
                mismatches.push_back(offset);
                offset++;
            }
            else if(isdigit(MDtag[j]))
            {
                digit_string.append(MDtag[j]);
            }
            else
            {
                fprintf(stderr, "Unexpected character found in MD tag for read %s: %c\n",
                    bam1_qname(mcur), MDtag[j]);
            }
        }

        char* quality = bam1_qual(mcur); /* This is not offset by 33 */
        for(j=0; j<mcur->core.l_qseq; j++)
        {
            if(mismatches.find(j) == mismatches.end()) /*It's a match */
            {
                Pcur *= pow( (double)10, (double)quality[j]) / (double)3;
            }
            else
            {
                Pcur *= 1 - pow( (double)10, (double)quality[j]);
            }
        }
        ret_val *= Pcur;
    }

    bam_destroy1(mcur);

    return ret_val;
}
void get_isoform_maps(bamFile refseq1, bamFile refseq2, bamFile fusions)
/*Get a map from an isoform string to a column in the matrix we'll eventually make. Also get a map from
 * isoform strings to the length of the isoform */
{
    bam_header_t* refseq1_header;
    bam_header_t* refseq2_header;
    bam_header_t* fusions_header;

    /*Initialize all of the headers*/
    refseq1_header = bam_header_init();
    refseq1_header = bam_header_read(refseq1);
    refseq2_header = bam_header_init();
    refseq2_header = bam_header_read(refseq2);
    fusions_header = bam_header_init();
    fusions_header = bam_header_read(fusions);

    /*First map isoform strings to lengths in columns*/
    int i;
    for(i=0; i<refseq1_header->n_targets; i++)
    {
        std::string isoform_string = std::string(refseq1_header->target_name[i]);
        int first_pipe = isoform_string.find_first_of("|");
        std::string transcript = isoform_string.substr(0, first_pipe);
        isoform_to_column[transcript] = i;
        isoform_lengths[transcript] = refseq1_header->target_len[i];
        header1_lengths.push_back(refseq1_header->target_len[i]);
    }
    int isoform_count = i;

    /*Then add isoforms that aren't present in refseq1*/
    std::map<std::string, int>::iterator isoform_it;
    for(i=0; i>refseq2_header->n_targets; i++)
    {
        std::string isoform_string = std::string(refseq2_header->target_name[i]);
        int first_pipe = isoform_string.find_first_of("|");
        std::string transcript = isoform_string.substr(0, first_pipe);
        isoform_it = isoform_to_column.find(transcript);
        if(isoform_it == isoform_to_column.end())
        {
            isoform_to_column[transcript] = isoform_count;
            isoform_lengths[transcript] = refseq2_header->target_len[i];
            isoform_count++;
        }
    }

    /*Now add fusion isoforms*/
    for(i=0; i<fusions_header->n_targets; i++)
    {
        std::string isoform_string = std::string(fusions_header->target_name[i]);
        int first_pipe = isoform_string.find_first_of("|");
        std::string transcript = isoform_string.substr(0, first_pipe);
        isoform_to_column[transcript] = isoform_count;
        isoform_count++;
        isoform_to_length[transcript] = fusions_header->target_len[i];
        /* We also need to update the map from refseq transcripts to fusions.
         * transcript for refseq mappings is like N[RM]_\d+
         * fusion transcripts are like F_N[MR]_\d+\.N[MR]_\d+_\d+_\d+ */
        char* transcript_c = transcript.c_str();
        char* left_side = strtok(transcript_c, '.'); /*Note that this changes the original */
        strtok(left_side, '_'); /* Now left_side is the US transcript */
        char right_side[75];
        size_t end_pos = strcspn(&transcript_c[3], "_") + 3;
        strncpy(right_side, transcript_c, end_pos);
        right_side[end_pos] = '\0';

        refseq_to_fusion[string(left_side)].push_back(transcript);
        refseq_to_fusion[string(right_side)].push_back(transcript);

        fusion_lengths.push_back(fusions_header->target_len[i]);
    }

    bam_header_destroy(refseq1_header);
    bam_header_destroy(refseq2_header);
    bam_header_destroy(fusions_header);
}



void initialize_mapping_matrix(comressed_matrix<double>& mapping_matrix, bamFile refseq1, bamFile refseq2,
        bamFile fusions)
{
    /* Note that all bamFiles should have read through their headers at this point,
     * and they should be sorted by read. */

    bam1_t* refseq1_mappings[500];
    bam1_t* refseq2_mappings[500];
    int i,j;
    for(i=0; i<500; i++)
    {
        refseq1_mappings[i] = bam_init1();
        refseq2_mappings[i] = bam_init1();
    }

    bam1_t* leftovers[2];
    leftovers[0] = bam_init1();
    leftovers[1] = bam_init1();
    bam_read1(refseq1, leftovers[0]);
    bam_read1(refseq2, leftovers[1]);

    bam1_t* current_record;
    current_record = bam_init1();
    int r1_mapping_count, r2_mapping_count;

    while(1)
    {

        /* First get all the mappings from the first read from refseq1 and refseq2 */
        r1_mapping_count = 1;
        bam_copy1(refseq1_mappings[0], leftovers[0]);
        while(1)
        {
            int read_bytes = bam_read1(refseq1, current_record);
            if(read_bytes == 0) break;
            if(strcmp(bam1_qname(current_record), bam1_qname(refseq1_mappings[0])) == 0)
            {
                bam_copy1(refseq1_mappings[r1_mapping_count], current_record);
                r1_mapping_count++;
            }
            else
            {
                bam_copy1(leftovers[0], current_record);
                break;
            }
        }

        r2_mapping_count = 1;
        bam_copy1(refseq2_mappings[0], leftovers[1]);
        while(1)
        {
            int read_bytes = bam_read1(refseq2, current_record);
            if(read_bytes == 0) break;
            if(strcmp(bam1_qname(current_record), bam1_qname(refseq2_mappings[0])) == 0)
            {
                bam_copy1(refseq2_mappings[r2_mapping_count], current_record);
                r2_mapping_count++;
            }
            else
            {
                bam_copy1(leftovers[1], current_record);
                break;
            }
        }

        /* If either end is unmapped, continue */
        if( (refseq1_mappings[0]->core.flag & BAM_FUNMAP) ||
            (refseq2_mappings[0]->core.flat & BAM_FUNMAP) )
            continue;

        /* Now see if there are good paired mappings between the reads */
        for(i=0; i<r1_mapping_count; i++)
        {
            for(j=0; j<r2_mapping_count; j++)
            {
                if( (refseq1_mappings[i]->core.tid == refseq2_mappings[j]->core.tid) && /* Same transcript */
                    (abs(refseq1_mappings[i]->core.pos - refseq2_mappings[j]->core.pos) < 2000) &&
                    ((refseq1_mappings[i]->core.flag & BAM_FREVERSE) !=
                     (refseq2_mappings[j]->core.flag & BAM_FREVERSE)) )
                {
                    printf("JACKPOT\n");
                }
            }
        }
    }
}






int main(int argc, char* argv[])
{
    bamFile refseq1, refseq2, fusions;
    if(argc != 4)
    {
        fprintf(stderr, "EM <refseq.1.bam> <refseq.2.bam> <fusions.bam>");
        return 1;
    }
    refseq1 = bam_open(argv[1], "r");
    refseq2 = bam_open(argv[2], "r");
    fusions = bam_open(argv[3], "r");

    EM_core(refseq1, refseq2, fusions);
    return 0;
}

