#ifndef SHORTFUSE_H
#define SHORTFUSE_H

#include <string>
#include <set>
#include <stdio.h>
#include "bam.h"

void get_seq_qual_strings(char* seq, char* qual, bam1_t* b)
{
    int i;
    for(i=0; i<b->core.l_qseq; i++)
    {
        int bit_base = bam1_seqi(bam1_seq(b), i);
        char base = '\0';
        switch (bit_base) {
            case 1:
                base = 'A';
                break;
            case 2:
                base = 'C';
                break;
            case 4:
                base = 'G';
                break;
            case 8:
                base = 'T';
                break;
            case 15:
                base = 'N';
        }
        seq[i] = base;
        qual[i] = bam1_qual(b)[i] + 33;
    }
    seq[i] = '\0';
    qual[i] = '\0';
}

void get_mismatch_indices(std::set<short int>& mismatches, bam1_t* b)
{
        char* MDtag = (char*)bam_aux_get(b, "MD");
        int offset = 0;
        std::string digit_string = "";
        for(int j=1; j<(int)strlen(MDtag); j++) /* The first char is always i */
        {
            char MDchar = (char)MDtag[j];
            if(isalpha(MDchar))
            {
                if(digit_string.length() > 0)
                {
                    offset += atoi(digit_string.c_str());
                    digit_string = "";
                }
                mismatches.insert(offset);
                offset++;
            }
            else if(isdigit(MDchar))
            {
                digit_string.append(&MDchar);
            }
            else
            {
                fprintf(stderr, "Unexpected character found in MD tag for read %s: %c\n",
                    bam1_qname(b), MDchar);
            }
        }
}

#endif
