#include <ctype.h>
#include <unistd.h>
#include "ShortFuse.h"
#include "bam.h"

#define MAX_MAPPING_COUNT 300

void bam_sift(bamFile in1, bamFile in2, char* out_prefix) {

    bam_header_t* header1;
    bam_header_t* header2;

    bam1_t* b1[MAX_MAPPING_COUNT];
    bam1_t* b2[MAX_MAPPING_COUNT];

    header1 = bam_header_init();
    header2 = bam_header_init();

    header1 = bam_header_read(in1);
    header2 = bam_header_read(in2);

    FILE* fastqfile;
    char fastqfilename[250];
    strcpy(fastqfilename, out_prefix);
    fastqfile = fopen(strcat(fastqfilename, ".discord.fastq"), "w");

    FILE* distprobfile;
    char distprobfilename[250];
    strcpy(distprobfilename, out_prefix);
    distprobfile = fopen(strcat(distprobfilename, ".distprob"), "w");

    int i;
    for(i=0; i<MAX_MAPPING_COUNT; i++){
        b1[i] = bam_init1();
        b2[i] = bam_init1();
    }

    bam1_t* curr1;
    bam1_t* curr2;
    bam1_t* leftover1;
    bam1_t* leftover2;
    curr1 = bam_init1();
    curr2 = bam_init1();
    leftover1 = bam_init1();
    leftover2 = bam_init1();

    leftover1->data_len = 0;
    leftover2->data_len = 0;

    int count1;
    int count2;

    int* mapping_distances = calloc(25000, sizeof(int));

    char current_qname[256] = "";

    int bam_eof = 0;
    while(!bam_eof) {
        int byte_count;
        count1 = 0;

        /*Get all BAM entries associated with the current qname*/
        if(leftover1->data_len != 0){
            bam_copy1(b1[0], leftover1);
            count1++;
            strncpy(current_qname, bam1_qname(b1[0]), strlen(bam1_qname(b1[0])) - 2);
            current_qname[strlen(bam1_qname(b1[0])) - 2] = '\0';
        }

        while(1) {
            char qname1[256] = "";
            byte_count = bam_read1(in1, curr1);
            if(byte_count <= 0) {
                bam_eof = 1;
                break;
            }
            strncpy(qname1, bam1_qname(curr1), strlen(bam1_qname(curr1)) - 2);
            qname1[strlen(bam1_qname(curr1)) - 2] = '\0';
            if(strlen(current_qname) == 0) {
                strncpy(current_qname, bam1_qname(curr1), strlen(bam1_qname(curr1)) - 2);
            }
            if(strcmp(current_qname, qname1) == 0) {
                bam_copy1(b1[count1], curr1);
                count1++;
            } else {
                bam_copy1(leftover1, curr1);
                break;
            }
        }

        count2 = 0;
        if(leftover2->data_len != 0){
            bam_copy1(b2[0], leftover2);
            count2++;
        }
        while(1) {
            char qname2[256] = "";
            byte_count = bam_read1(in2, curr2);
            if(byte_count <= 0) {
                bam_eof = 1;
                break;
            }
            strncpy(qname2, bam1_qname(curr2), strlen(bam1_qname(curr2)) - 2);
            qname2[strlen(bam1_qname(curr2)) - 2] = '\0';
            if(strcmp(current_qname, qname2) == 0) {
                bam_copy1(b2[count2], curr2);
                count2++;
            } else {
                bam_copy1(leftover2, curr2);
                break;
            }
        }

        /*See if either end is unmapped. If so, continue*/
        if(b1[0]->core.flag & BAM_FUNMAP || b2[0]->core.flag & BAM_FUNMAP) continue;

        /*Get the names of the mapped transcripts for each end of the read*/
        char** transcript_names1 = malloc(256 * sizeof(char*));
        char** gene_names1 = malloc(256 * sizeof(char*));
        for(i=0; i<count1; i++){
            transcript_names1[i] = malloc(64 * sizeof(char));
            gene_names1[i] = malloc(64 * sizeof(char));
            char* bam_mapping = header1->target_name[b1[i]->core.tid];
            char mapping[256];
            strcpy(mapping, bam_mapping);
            char* tok = strtok(mapping, "|");
            strcpy(transcript_names1[i], tok);
            tok = strtok(NULL, "|");
            strcpy(gene_names1[i], tok);
        }

        char** transcript_names2 = malloc(256 * sizeof(char*));
        char** gene_names2 = malloc(256 * sizeof(char*));
        for(i=0; i<count2; i++){
            transcript_names2[i] = malloc(64 * sizeof(char));
            gene_names2[i] = malloc(64 * sizeof(char));
            char* bam_mapping = header2->target_name[b2[i]->core.tid];
            char mapping[256];
            strcpy(mapping, bam_mapping);
            char* tok = strtok(mapping, "|");
            strcpy(transcript_names2[i], tok);
            tok = strtok(NULL, "|");
            strcpy(gene_names2[i], tok);
        }

        /*See if there is any overlap between the gene names*/
        int j;
        int overlap = 0;
        for(i=0; i<count1; i++){
            for(j=0; j<count2; j++){
                if(strcmp(gene_names1[i], gene_names2[j]) == 0){
                    overlap = 1;
                    break;
                }
            }
        }

        if(overlap) {
            for(i=0; i<count1; i++){
                for(j=0; j<count2; j++){
                    if(strcmp(transcript_names1[i], transcript_names2[j]) == 0){
                        if((b1[i]->core.flag & BAM_FREVERSE) != (b2[j]->core.flag & BAM_FREVERSE)){
                            int mapping_distance = abs(b1[i]->core.pos - b2[j]->core.pos);
                            if(mapping_distance < 25000) mapping_distances[mapping_distance]++;
                        }
                    }
                }
            }
        } else { /*Only discordant mappings*/
            int write_to_fastq = 0;
            for(i=0; i<count1; i++){
                for(j=0; j<count2; j++){
                    if((b1[i]->core.flag & BAM_FREVERSE) == (b2[j]->core.flag & BAM_FREVERSE)) continue;
                    int mismatch_count1 = 0;
                    int mismatch_count2 = 0;
                    int mismatches1[50];
                    int mismatches2[50];
                    char* mdtag = (char*)bam_aux_get(b1[i], "MD");
                    int md_i;
                    for(md_i=1; md_i<strlen(mdtag); md_i++){
                        if(isalpha(mdtag[md_i])) {
                            mismatches1[mismatch_count1] = md_i;
                            mismatch_count1++;
                        }
                    }

                    mdtag = (char*)bam_aux_get(b2[j], "MD");
                    for(md_i=1; md_i<strlen(mdtag); md_i++){
                        if(isalpha(mdtag[md_i])) {
                            mismatches2[mismatch_count2] = md_i;
                            mismatch_count2++;
                        }
                    }

                    if(mismatch_count1 > 2 || mismatch_count2 > 2) continue;
                    write_to_fastq = 1;
                    char base_read_id[128];
                    strncpy(base_read_id, bam1_qname(b1[i]), strlen(bam1_qname(b1[i])) - 2);
                    base_read_id[strlen(bam1_qname(b1[i])) - 2] = '\0';
                    printf("%s\t", base_read_id);
                    if(((b1[i]->core.flag & BAM_FREVERSE) == 0) && ((b2[j]->core.flag & BAM_FREVERSE) != 0)){
                        /*1 is forward strand*/
                        printf("%s\t%s\t%s\t%s\t%d\t%d\t%d\t%d\t",
                                header1->target_name[b1[i]->core.tid],
                                header2->target_name[b2[j]->core.tid],
                                gene_names1[i],
                                gene_names2[j],
                                b1[i]->core.pos,
                                b2[j]->core.pos,
                                b1[i]->core.l_qseq,
                                b2[j]->core.l_qseq);
                        for(md_i=0; md_i<mismatch_count1; md_i++){
                            printf("%d,", mismatches1[md_i]);
                        }
                        printf("\t");
                        for(md_i=0; md_i<mismatch_count2; md_i++){
                            printf("%d,", mismatches2[md_i]);
                        }
                        printf("\n");
                    } else
                    if(((b1[i]->core.flag & BAM_FREVERSE) != 0) && ((b2[j]->core.flag & BAM_FREVERSE) == 0)){
                        /*1 is forward strand*/
                        printf("%s\t%s\t%s\t%s\t%d\t%d\t%d\t%d\t",
                                header2->target_name[b2[j]->core.tid],
                                header1->target_name[b1[i]->core.tid],
                                gene_names2[j],
                                gene_names1[i],
                                b2[j]->core.pos,
                                b1[i]->core.pos,
                                b2[j]->core.l_qseq,
                                b1[i]->core.l_qseq);
                        for(md_i=0; md_i<mismatch_count2; md_i++){
                            printf("%d,", mismatches2[md_i]);
                        }
                        printf("\t");
                        for(md_i=0; md_i<mismatch_count1; md_i++){
                            printf("%d,", mismatches1[md_i]);
                        }
                        printf("\n");
                    }
                }
            }
            if(write_to_fastq){
                char seq[200];
                char qual[200];
                for(i=0; i<b1[0]->core.l_qseq; i++){
                    int bit_base = bam1_seqi(bam1_seq(b1[0]), i);
                    char base = '\0';
                    switch (bit_base) {
                        case 1:
                            base = 'A';
                            break;
                        case 2:
                            base = 'C';
                            break;
                        case 4:
                            base = 'G';
                            break;
                        case 8:
                            base = 'T';
                            break;
                        case 15:
                            base = 'N';
                    }
                    seq[i] = base;
                    qual[i] = bam1_qual(b1[0])[i] + 33;
                }
                seq[i] = '\0';
                qual[i] = '\0';

                fprintf(fastqfile, "@%s\n%s\n+\n%s\n", bam1_qname(b1[0]),
                        seq, qual);
                for(i=0; i<b2[0]->core.l_qseq; i++){
                    int bit_base = bam1_seqi(bam1_seq(b2[0]), i);
                    char base = '\0';
                    switch (bit_base) {
                        case 1:
                            base = 'A';
                            break;
                        case 2:
                            base = 'C';
                            break;
                        case 4:
                            base = 'G';
                            break;
                        case 8:
                            base = 'T';
                            break;
                        case 15:
                            base = 'N';
                    }
                    seq[i] = base;
                    qual[i] = bam1_qual(b2[0])[i] + 33;
                }
                seq[i] = '\0';
                qual[i] = '\0';
                fprintf(fastqfile, "@%s\n%s\n+\n%s\n", bam1_qname(b2[0]),
                        seq, qual);
            }
        }
        for(i=0; i<count1; i++){
            free(transcript_names1[i]);
            free(gene_names1[i]);
        }
        free(transcript_names1);
        free(gene_names1);

        for(i=0; i<count2; i++){
            free(transcript_names2[i]);
            free(gene_names2[i]);
        }
        free(transcript_names2);
        free(gene_names2);
    }

    int mapping_distance_sum = 0;
    for(i=0;i<25000;i++){
        if(mapping_distances[i] == 0) mapping_distances[i] = 1;
        mapping_distance_sum += mapping_distances[i];
    }

    for(i=0;i<25000;i++){
        fprintf(distprobfile, "%d\t%f\n", i, (double)mapping_distances[i]/mapping_distance_sum);
    }

}

int main(int argc, char* argv[]){

    bamFile in1, in2;
    if(argc != 4){
        fprintf(stderr, "sift <in.1.bam> <in.2.bam> <out.prefix>");
        return 1;
    }
    in1 = bam_open(argv[1], "r");
    in2 = bam_open(argv[2], "r");
    bam_sift(in1, in2, argv[3]);
    return 0;
}
